@@ -537,7 +537,10 @@ class Application(object):
         for attr in dir(obj):
             if attr.endswith('__'):
                 continue
-            setattr(obj, attr, count(getattr(obj, attr)))
+            value = getattr(obj, attr)
+            if not callable(value):
+                continue
+            setattr(obj, attr, count(value))
         obj.counter = counter
         return obj
 
@@ -601,22 +604,55 @@ class Application(object):
     def kwargs(self, **kwargs):
         return kwargs
 
-    @zeronimo.rpc(manual_ack=True)
-    def maybe_reject(self, ack, x, y):
-        ack(not getattr(self.maybe_reject, 'reject', False))
-        return x + y
-
-    @zeronimo.rpc(manual_ack=True)
-    def iter_maybe_reject(self, ack, x, y):
-        self.iter_maybe_reject.final_state = None
-        ack(not getattr(self.iter_maybe_reject, 'reject', False))
-        self.iter_maybe_reject.final_state = (False, x, y)
-        yield x
-        yield y
-        self.iter_maybe_reject.final_state = (True, x, y)
-
     def is_remote(self):
         return False
     @zeronimo.rpc('is_remote')
     def _rpc_is_remote(self):
         return True
+
+    @zeronimo.rpc(pass_call=True)
+    def hints(self, call):
+        return call.hints
+
+    @zeronimo.rpc
+    def reject_by_hints(self):
+        return 'accepted'
+
+    @reject_by_hints.reject_if
+    def _reject_by_hints(self, call, topics):
+        assert isinstance(self, Application)
+        return 'reject' in call.hints
+
+    @zeronimo.rpc
+    def reject_by_hints_staticmethod(self):
+        return 'accepted'
+
+    @reject_by_hints_staticmethod.reject_if
+    @staticmethod
+    def _reject_by_hints_staticmethod(call, topics):
+        return 'reject' in call.hints
+
+    @zeronimo.rpc
+    def reject_by_hints_classmethod(self):
+        return 'accepted'
+
+    @reject_by_hints_classmethod.reject_if
+    @classmethod
+    def _reject_by_hints_classmethod(cls, call, topics):
+        assert isinstance(cls, type)
+        return 'reject' in call.hints
+
+    odd = False
+
+    @zeronimo.rpc
+    def reject_by_odd_even(self):
+        return 'accepted'
+
+    @reject_by_odd_even.reject_if
+    def _reject_by_odd_even(self, call, topics):
+        reject = self.odd
+        self.odd = not self.odd
+        return reject
+
+    def raise_remote_value_error(self, value='zeronimo'):
+        raise zeronimo.RemoteException.compose(ValueError)(value)
