@@ -278,41 +278,6 @@ def test_error_on_reject_if(worker, collector, pub, topic):
     assert 'ZeroDivisionError' in str(w[0].message)
 
 
-def test_manual_ack(worker1, worker2, collector, push):
-    # Workers wrap own object.
-    worker1.app = Application()
-    worker2.app = Application()
-    errors = []
-    worker1.exception_handler = lambda w, e: errors.append(e)
-    # Any worker accepts.
-    customer = zeronimo.Customer(push, collector)
-    assert customer.call('maybe_reject', 1, 2).get() == 3
-    # worker1 will reject.
-    worker1.app.maybe_reject.reject = True
-    # But worker2 still accepts.
-    assert customer.call('maybe_reject', 3, 4).get() == 7
-    assert not errors
-    # worker2 will also reject.
-    worker2.app.maybe_reject.reject = True
-    # All workers reject.
-    customer.timeout = 0.1
-    with pytest.raises(Rejected):
-        customer.call('maybe_reject', 5, 6)
-    # A generator marked as manual_rpc=True.
-    worker1.app.iter_maybe_reject.reject = True
-    gen = customer.call('iter_maybe_reject', 7, 8).get()
-    assert next(gen) == 7
-    assert next(gen) == 8
-    assert worker2.app.iter_maybe_reject.final_state == (True, 7, 8)
-    # All generators marked as manual_rpc=True.
-    worker2.app.iter_maybe_reject.reject = True
-    with pytest.raises(Rejected):
-        customer.call('iter_maybe_reject', 9, 10)
-    assert worker1.app.iter_maybe_reject.final_state is None
-    assert worker2.app.iter_maybe_reject.final_state is None
-    assert not errors
-
-
 def test_max_retries(worker, collector, push, monkeypatch):
     def stop_accepting():
         monkeypatch.setattr(worker.greenlet_group, 'full', lambda: True)
@@ -530,11 +495,13 @@ def test_malformed_message(worker, push):
     push.send_multipart([zeronimo.messaging.SEAM, 'x', 'y'])
     expect('TypeError')
     push.send_multipart([zeronimo.messaging.SEAM, 'x', 'y', 'z'])
-    expect('UnpicklingError')
+    expect('AttributeError')
     push.send_multipart([zeronimo.messaging.SEAM, 'x', 'y', 'z', 'w'])
     expect('AttributeError')
     push.send_multipart([zeronimo.messaging.SEAM, 'x', 'y', 'z',
                          'c__main__\nNOTEXIST\np0\n.'])
+    expect('UnpicklingError')
+    push.send_multipart([zeronimo.messaging.SEAM, 'zeronimo', 'y', 'z', 'w'])
     with warnings.catch_warnings(record=True) as w:
         gevent.sleep(0.1)
     assert len(w) == len(expectations)
@@ -1013,6 +980,130 @@ def test_unicode(worker, push, collector):
     assert customer.call(u'zeronimo').get() == 'zeronimo'
 
 
+def test_hints(worker, pub, collector, topic):
+    fanout = zeronimo.Fanout(pub, collector)
+    worker.reject_if = lambda call, topics: 'reject' in call.hints
+    r = fanout.emit(topic, ['hello', 'world'], 'zeronimo')
+    assert get_results(r) == ['zeronimo']
+    r = fanout.emit(topic, ['hello', 'world'], 'zeronimo')
+    assert get_results(r) == ['zeronimo']
+    r = fanout.emit(topic, ['reject'], 'zeronimo')
+    assert get_results(r) == []
+    r = fanout.emit(topic, ['foo', 'reject', 'bar'], 'zeronimo')
+    assert get_results(r) == []
+    r = fanout.emit(topic, ['hello', 'world'], 'hints')
+    assert get_results(r) == [('hello', 'world')]
+
+
+def test_bound_hints(worker, pub, collector, topic):
+    fanout = zeronimo.Fanout(pub, collector, hints=['reject'])
+    r = fanout.emit(topic, ['hello', 'world'], 'zeronimo')
+    assert get_results(r) == ['zeronimo']
+    worker.reject_if = lambda call, topics: 'reject' in call.hints
+    r = fanout.emit(topic, ['hello', 'world'], 'zeronimo')
+    assert get_results(r) == []
+
+
+def test_raw(worker, push, pub, collector, topic):
+    customer = zeronimo.Customer(push, collector)
+    fanout = zeronimo.Fanout(pub, collector)
+    payload = zeronimo.messaging.PACK(((), {'hello': 'world'}))
+    assert customer.call_raw('kwargs', payload).get() == {'hello': 'world'}
+    assert \
+        get_results(fanout.emit_raw(topic, 'kwargs', payload)) == \
+        [{'hello': 'world'}]
+    with pytest.raises(TypeError):
+        customer.call_raw('kwargs')
+    with pytest.raises(TypeError):
+        customer.call_raw('kwargs', payload=payload)
+    with pytest.raises(TypeError):
+        customer.call_raw('kwargs', payload, payload)
+
+
+def test_specific_reject_if(worker, push, collector):
+    customer = zeronimo.Customer(push, collector, timeout=0.1)
+    assert customer.call('reject_by_hints').get() == 'accepted'
+    assert customer.call('reject_by_hints_staticmethod').get() == 'accepted'
+    assert customer.call('reject_by_hints_classmethod').get() == 'accepted'
+    unpack_called = []
+    def unpack(*args, **kwargs):
+        unpack_called.append(True)
+        return zeronimo.messaging.UNPACK(*args, **kwargs)
+    worker.unpack = unpack
+    assert not unpack_called
+    with pytest.raises(Rejected):
+        customer.call(['reject'], 'reject_by_hints').wait()
+    assert not unpack_called
+    with pytest.raises(Rejected):
+        customer.call(['reject'], 'reject_by_hints_staticmethod').wait()
+    assert not unpack_called
+    with pytest.raises(Rejected):
+        customer.call(['reject'], 'reject_by_hints_classmethod').wait()
+    assert not unpack_called
+    with pytest.raises(Rejected):
+        customer.call('reject_by_odd_even').wait()
+    assert unpack_called
+
+
+def test_require_rpc_specs(worker, push, collector):
+    worker.require_rpc_specs = True
+    customer = zeronimo.Customer(push, collector, timeout=0.1)
+    with pytest.raises(Rejected):
+        customer.call('add', 0, 42).wait()
+    assert customer.call('zeronimo').get() == 'zeronimo'
+    with pytest.raises(Rejected):
+        customer.call('_zeronimo').wait()
+
+
+def test_trace(worker, push, collector):
+    # Log traced events.
+    traced = []
+    def trace(method, args):
+        traced.append((method,) + args)
+    collector.trace = trace
+    customer = zeronimo.Customer(push, collector, timeout=0.1)
+    customer.call('zeronimo').get()
+    assert len(traced) == 3
+    call_id = traced[0][1]
+    assert traced[0] == (0, call_id, customer, 'zeronimo', (), {})
+    assert traced[1][:2] == (zeronimo.messaging.ACCEPT, call_id)
+    reply_id = traced[1][2]
+    assert traced[1][3] == worker.info
+    assert traced[2] == (zeronimo.messaging.RETURN,
+                         call_id, reply_id, 'zeronimo')
+    # Trace RPC execution time.
+    def trace_time(mem, method, args):
+        if method == 0:
+            mem.called_at = time.time()
+        elif method == zeronimo.messaging.RETURN:
+            mem.returned_at = time.time()
+    mem = type('', (), {})()
+    collector.trace = lambda *x: trace_time(mem, *x)
+    customer.call('sleep', 0.1).get()
+    assert 0.1 <= mem.returned_at - mem.called_at < 0.2
+
+
+def test_raise_remote_exception(worker, push, collector):
+    customer = zeronimo.Customer(push, collector, timeout=0.1)
+    try:
+        customer.call('raise_remote_value_error', 'zeronimo').get()
+    except BaseException as exc:
+        pass
+    else:
+        assert False, 'not raised'
+    assert isinstance(exc, ValueError)
+    assert isinstance(exc, zeronimo.RemoteException)
+    assert exc.args[0] == 'zeronimo'
+
+
+def test_parse():
+    f = zeronimo.messaging.parse
+    header, payload, topics = f(['hello', 'world', '\xff', '1', '2', '3', '4'])
+    assert header == ['1', '2', '3']
+    assert payload == '4'
+    assert topics == ['hello', 'world']
+
+
 # catch leaks
 
 
