@@ -3,7 +3,7 @@
    zeronimo.core
    ~~~~~~~~~~~~~
 
-   :copyright: (c) 2013-2016 by Heungsub Lee
+   :copyright: (c) 2013-2017 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
 
 """
@@ -11,7 +11,7 @@ from __future__ import absolute_import
 
 from collections import Iterator
 from contextlib import contextmanager
-from functools import partial
+from StringIO import StringIO
 import sys
 import traceback
 from warnings import warn
@@ -26,15 +26,16 @@ except ImportError:
     uuid4_bytes = lambda: uuid.uuid4().get_bytes()
 import zmq.green as zmq
 
-from .application import default_rpc_spec, rpc_table
-from .exceptions import (
-    EmissionError, MalformedMessage, Reject, Rejected, TaskClosed, Undelivered,
+from zeronimo.application import NULL_RPC_SPEC, rpc_spec_table
+from zeronimo.exceptions import (
+    EmissionError, MalformedMessage, Rejected, TaskClosed, Undelivered,
     WorkerNotFound)
-from .helpers import class_name, eintr_retry_zmq as safe, Flag
-from .messaging import (
+from zeronimo.helpers import (
+    class_name, eintr_retry_zmq as safe, FALSE_RETURNER, repr_socket)
+from zeronimo.messaging import (
     ACCEPT, ACK, BREAK, Call, PACK, RAISE, recv, REJECT, Reply, RETURN, send,
     UNPACK, YIELD)
-from .results import RemoteResult
+from zeronimo.results import RemoteException, RemoteResult
 
 
 __all__ = ['Worker', 'Customer', 'Fanout', 'Collector']
@@ -79,8 +80,8 @@ class Background(object):
 
     def __call__(self):
         # should be implemented by subclass.
-        raise NotImplementedError('{0} has no __call__ implementation'
-                                  ''.format(class_name(self)))
+        raise NotImplementedError('%s has no __call__ implementation'
+                                  '' % class_name(self))
 
     def run(self):
         try:
@@ -94,7 +95,7 @@ class Background(object):
         if self.is_running():
             if silent:
                 return
-            raise RuntimeError('{0} already running'.format(class_name(self)))
+            raise RuntimeError('%s already running' % class_name(self))
         self.greenlet = self.greenlet_class.spawn(self.run)
         self.greenlet.join(0)
         return self.greenlet
@@ -103,12 +104,12 @@ class Background(object):
         if not self.is_running():
             if silent:
                 return
-            raise RuntimeError('{0} not running'.format(class_name(self)))
+            raise RuntimeError('%s not running' % class_name(self))
         self.greenlet.kill(block=True)
 
     def wait(self, timeout=None):
         if not self.is_running():
-            raise RuntimeError('{0} not running'.format(class_name(self)))
+            raise RuntimeError('%s not running' % class_name(self))
         self.greenlet.join(timeout)
 
     def is_running(self):
@@ -141,24 +142,6 @@ def default_malformed_message_handler(worker, exc_info, message_parts):
     warn('<%s> occurred by %r' % (exc_str, message_parts), MalformedMessage)
 
 
-def _ack(worker, reply_socket, channel, call, acked,
-         accept=True, silent=False):
-    if not reply_socket:
-        return
-    elif acked:
-        if silent:
-            return
-        raise RuntimeError('already acknowledged')
-    acked(True)
-    if accept:
-        worker.accept(reply_socket, channel)
-        return
-    __, __, topics = channel
-    worker.reject(reply_socket, call.call_id, topics)
-    if not silent:
-        raise Reject
-
-
 class Worker(Background):
     """A worker runs an RPC service of an object through ZeroMQ sockets:
 
@@ -189,7 +172,8 @@ class Worker(Background):
                  info=None, greenlet_group=None,
                  exception_handler=default_exception_handler,
                  malformed_message_handler=default_malformed_message_handler,
-                 reject_if=(lambda *__: False), pack=PACK, unpack=UNPACK):
+                 reject_if=FALSE_RETURNER, require_rpc_specs=False,
+                 pack=PACK, unpack=UNPACK):
         super(Worker, self).__init__()
         self.app = app
         self.sockets = sockets
@@ -201,6 +185,7 @@ class Worker(Background):
         self.exception_handler = exception_handler
         self.malformed_message_handler = malformed_message_handler
         self.reject_if = reject_if
+        self.require_rpc_specs = require_rpc_specs
         self.pack = pack
         self.unpack = unpack
 
@@ -211,12 +196,12 @@ class Worker(Background):
     @app.setter
     def app(self, app):
         self._app = app
-        self.rpc_table = rpc_table(app)
+        self.rpc_spec_table = rpc_spec_table(app)
 
     @app.deleter
     def app(self):
         del self._app
-        del self.rpc_table
+        del self.rpc_spec_table
 
     @property
     def obj(self):
@@ -234,9 +219,18 @@ class Worker(Background):
         def accept(socket, call, args, kwargs, topics):
             group.spawn(self.work, socket, call, args, kwargs, topics)
             group.join(0)
-        def reject(socket, (__, call_id, reply_to), topics):
+        def reject(socket, call, topics):
+            __, call_id, reply_to, __ = call
             reply_socket, topics = self.replier(socket, topics, reply_to)
             self.reject(reply_socket, call_id, topics)
+        def reject_if(call, topics):
+            if self.reject_if(call, topics):
+                return True
+            try:
+                __, rpc_spec = self.find_call_target(call)
+            except KeyError:
+                return True
+            return rpc_spec.reject_if.__get__(self.app)(call, topics)
         try:
             while True:
                 for socket, event in safe(poller.poll):
@@ -244,9 +238,8 @@ class Worker(Background):
                     del msgs[:]
                     try:
                         header, payload, topics = recv(socket, capture=capture)
-                        call = Call(*header)
-                        if group.full() or self.reject_if(call, topics):
-                            # Reject the call if it should.
+                        call = Call(*(header[:3] + [tuple(header[3:])]))
+                        if group.full() or reject_if(call, topics):
                             reject(socket, call, topics)
                             continue
                         args, kwargs = self.unpack(payload)
@@ -284,20 +277,16 @@ class Worker(Background):
         else:
             channel = (None, None, None)
         f, rpc_spec = self.find_call_target(call)
-        acked = Flag()
-        ack = partial(_ack, self, reply_socket, channel, call, acked)
-        if not rpc_spec.manual_ack:
-            # Acknowledge automatically.
-            ack()
+        if rpc_spec.reject_if.__get__(self.app)(call, topics):
+            reply_socket and self.reject(reply_socket, call.call_id, topics)
+            return
+        reply_socket and self.accept(reply_socket, channel)
         success = False
         with self.catch_exceptions():
             try:
-                val = self.call(call, args, kwargs, ack, f, rpc_spec)
-            except Reject:
-                return
+                val = self.call(call, args, kwargs, f, rpc_spec)
             except:
                 exc_info = sys.exc_info()
-                ack(accept=False, silent=True)
                 self.raise_(reply_socket, channel, exc_info)
                 raise exc_info[0], exc_info[1], exc_info[2]
             success = True
@@ -311,35 +300,32 @@ class Worker(Background):
                     try:
                         val = next(vals)
                     except StopIteration:
-                        ack(accept=True, silent=True)
+                        pass
                     else:
-                        ack(accept=True, silent=True)
                         self.send_reply(reply_socket, YIELD, val, *channel)
                         for val in vals:
                             self.send_reply(reply_socket, YIELD, val, *channel)
                     self.send_reply(reply_socket, BREAK, None, *channel)
-                except Reject:
-                    return
                 except:
                     exc_info = sys.exc_info()
-                    ack(accept=False, silent=True)
                     self.raise_(reply_socket, channel, exc_info)
                     raise exc_info[0], exc_info[1], exc_info[2]
         else:
-            ack(accept=True, silent=True)
             self.send_reply(reply_socket, RETURN, val, *channel)
 
     def find_call_target(self, call):
         try:
-            return self.rpc_table[call.name]
+            return self.rpc_spec_table[call.name]
         except KeyError:
-            return getattr(self.app, call.name), default_rpc_spec
+            if self.require_rpc_specs:
+                raise
+            return getattr(self.app, call.name), NULL_RPC_SPEC
 
-    def call(self, call, args, kwargs, ack, f=None, rpc_spec=None):
+    def call(self, call, args, kwargs, f=None, rpc_spec=None):
         if f is None and rpc_spec is None:
             f, rpc_spec = self.find_call_target(call)
-        if rpc_spec.manual_ack:
-            args = (ack,) + args
+        if rpc_spec.pass_call:
+            args = (call,) + args
         return f(*args, **kwargs)
 
     def accept(self, reply_socket, channel):
@@ -360,6 +346,8 @@ class Worker(Background):
         exc_type, exc, tb = exc_info
         while tb.tb_next is not None:
             tb = tb.tb_next
+        if issubclass(exc_type, RemoteException):
+            exc_type = exc_type.exc_type
         filename, lineno = tb.tb_frame.f_code.co_filename, tb.tb_lineno
         val = (exc_type, str(exc), filename, lineno)
         try:
@@ -409,7 +397,16 @@ class Worker(Background):
         return self.greenlet_group.join(timeout, raise_error)
 
     def __repr__(self):
-        return '<{0} info={1!r}>'.format(class_name(self), self.info)
+        buf = StringIO()
+        buf.write('<%s' % class_name(self))
+        if self.info is not None:
+            buf.write(' info=%r' % self.info)
+        sockets = ', '.join(repr_socket(s) for s in self.sockets)
+        buf.write(' sockets=[%s]' % sockets)
+        if self.reply_socket is not None:
+            buf.write(' reply_socket=%s' % repr_socket(self.reply_socket))
+        buf.write('>')
+        return buf.getvalue()
 
 
 class _Caller(object):
@@ -422,25 +419,30 @@ class _Caller(object):
 
     socket = None
     collector = None
+    hints = ()
     pack = None
 
-    def __init__(self, socket, collector=None, timeout=None, pack=PACK):
+    def __init__(self, socket, collector=None, timeout=None,
+                 hints=(), pack=PACK):
         self.socket = socket
         self.collector = collector
+        self.hints = hints
         self.pack = pack
         if timeout is not None:
             self.timeout = timeout
 
-    def _call_nowait(self, name, args, kwargs, topics=()):
+    def _call_nowait(self, hints, name, args, kwargs, topics=(), raw=False):
         header = [name.encode(ENCODING), '', NO_REPLY]
-        payload = self.pack((args, kwargs))
+        header.extend(self.hints)
+        header.extend(hints)
+        payload = self._pack(args, kwargs, raw)
         try:
             safe(send, self.socket, header, payload, topics, zmq.NOBLOCK)
         except zmq.Again:
             pass  # ignore.
 
-    def _call(self, name, args, kwargs, topics=(),
-              limit=None, retry=False, max_retries=None):
+    def _call_wait(self, hints, name, args, kwargs, topics=(), raw=False,
+                   limit=None, retry=False, max_retries=None):
         """Allocates a call id and emit."""
         col = self.collector
         if not col.is_running():
@@ -449,22 +451,53 @@ class _Caller(object):
         reply_to = (DUPLEX if self.socket is col.socket else col.topic)
         # Normal tuple is faster than namedtuple.
         header = [name.encode(ENCODING), call_id, reply_to]
-        payload = self.pack((args, kwargs))
+        header.extend(self.hints)
+        header.extend(hints)
+        payload = self._pack(args, kwargs, raw)
         # Use short names.
         def send_call():
             try:
                 safe(send, self.socket, header, payload, topics, zmq.NOBLOCK)
             except zmq.Again:
                 raise Undelivered('emission was not delivered')
-        col.prepare(call_id)
+        col.prepare(call_id, self, name, args, kwargs)
         send_call()
         return col.establish(call_id, self.timeout, limit,
                              send_call if retry else None,
                              max_retries=max_retries)
 
+    def _pack(self, args, kwargs, raw=False):
+        if raw:
+            if kwargs or len(args) != 1:
+                raise TypeError('too many arguments')
+            return args[0]
+        return self.pack((args, kwargs))
+
     def close(self):
         self.socket.close()
 
+    def __repr__(self):
+        buf = StringIO()
+        buf.write('<%s socket=%s' % (class_name(self),
+                                     repr_socket(self.socket)))
+        if self.collector is not None:
+            buf.write(' collector=%r' % self.collector)
+        if self.timeout is not None:
+            buf.write(' timeout=%.3f' % self.timeout)
+        if self.hints:
+            buf.write(' hints=%r' % (tuple(self.hints),))
+        buf.write('>')
+        return buf.getvalue()
+
+
+def split_call_args(args, start=0):
+    if isinstance(args[start], basestring):
+        name, args = args[start], args[start + 1:]
+        hints = ()
+    else:
+        hints, name, args = args[start], args[start + 1], args[start + 2:]
+    return hints, name, args
+
 
 class Customer(_Caller):
     """A customer is a caller that sends an RPC call to one of workers at once.
@@ -479,15 +512,22 @@ class Customer(_Caller):
             self.max_retries = max_retries
         super(Customer, self).__init__(*args, **kwargs)
 
-    def call(self, *args, **kwargs):
-        name, args = args[0], args[1:]
+    def _call(self, raw, *args, **kwargs):
+        hints, name, args = split_call_args(args)
         if self.collector is None:
-            self._call_nowait(name, args, kwargs)
+            self._call_nowait(hints, name, args, kwargs, raw=raw)
             return
-        results = self._call(name, args, kwargs, limit=1,
-                             retry=True, max_retries=self.max_retries)
+        results = self._call_wait(hints, name, args, kwargs,
+                                  raw=raw, limit=1, retry=True,
+                                  max_retries=self.max_retries)
         return results[0]
 
+    def call(self, *args, **kwargs):
+        return self._call(False, *args, **kwargs)
+
+    def call_raw(self, *args, **kwargs):
+        return self._call(True, *args, **kwargs)
+
 
 class Fanout(_Caller):
     """A fanout is a caller that sends an RPC call to all workers subscribing
@@ -507,45 +547,54 @@ class Fanout(_Caller):
     drop_if = None
 
     def __init__(self, *args, **kwargs):
-        self.drop_if = kwargs.pop('drop_if', None)
+        self.drop_if = kwargs.pop('drop_if', FALSE_RETURNER)
         super(Fanout, self).__init__(*args, **kwargs)
 
-    def emit(self, *args, **kwargs):
+    def _emit(self, raw, *args, **kwargs):
         topic = args[0]
-        if self.drop_if is not None and self.drop_if(topic):
+        if self.drop_if(topic):
             # Drop the call without emission.
             return None if self.collector is None else []
-        name, args = args[1], args[2:]
+        hints, name, args = split_call_args(args, start=1)
         topics = (topic,) if topic else ()
         if self.collector is None:
-            self._call_nowait(name, args, kwargs, topics)
+            self._call_nowait(hints, name, args, kwargs, topics, raw=raw)
             return
         try:
-            return self._call(name, args, kwargs, topics)
+            return self._call_wait(hints, name, args, kwargs, topics, raw=raw)
         except EmissionError:
             return []
 
+    def emit(self, *args, **kwargs):
+        return self._emit(False, *args, **kwargs)
+
+    def emit_raw(self, *args, **kwargs):
+        return self._emit(True, *args, **kwargs)
+
 
 class Collector(Background):
     """A collector receives RPC results from workers."""
 
     socket = None
     topic = None
+    trace = None
     unpack = None
 
-    def __init__(self, socket, topic='', unpack=UNPACK):
+    def __init__(self, socket, topic='', trace=None, unpack=UNPACK):
         super(Collector, self).__init__()
         self.socket = socket
         self.topic = topic
+        self.trace = trace
         self.unpack = unpack
         self.results = {}
         self.result_queues = {}
 
-    def prepare(self, call_id):
+    def prepare(self, call_id, caller, name, args, kwargs):
         if call_id in self.results:
-            raise KeyError('call-{0} already prepared'.format(call_id))
+            raise KeyError('call-%s already prepared' % call_id)
         self.results[call_id] = {}
         self.result_queues[call_id] = Queue()
+        self.trace and self.trace(0, (call_id, caller, name, args, kwargs))
 
     def establish(self, call_id, timeout, limit=None,
                   retry=None, max_retries=None):
@@ -576,8 +625,9 @@ class Collector(Background):
             self.remove_result_queue(call_id)
         if not results:
             if rejected:
-                raise Rejected('{0} workers rejected'.format(rejected)
-                               if rejected != 1 else 'A worker rejected')
+                raise Rejected('%d workers rejected' % rejected
+                               if rejected != 1 else
+                               'A worker rejected')
             else:
                 raise WorkerNotFound('failed to find worker')
         return results
@@ -597,10 +647,12 @@ class Collector(Background):
             except:
                 # TODO: warn MalformedMessage
                 continue
-            method, call_id, reply_id = header
-            reply = Reply(ord(method), call_id, reply_id)
+            method, call_id, task_id = header
+            method = ord(method)
+            reply = Reply(method, call_id, task_id)
             value = self.unpack(payload)
-            del header, payload, method, call_id, reply_id
+            self.trace and self.trace(method, (call_id, task_id, value))
+            del header, payload, method, call_id, task_id
             try:
                 self.dispatch_reply(reply, value)
             except KeyError:
@@ -646,3 +698,12 @@ class Collector(Background):
     def close(self):
         super(Collector, self).close()
         self.socket.close()
+
+    def __repr__(self):
+        buf = StringIO()
+        buf.write('<%s socket=%s' % (class_name(self),
+                                     repr_socket(self.socket)))
+        if self.topic is not None:
+            buf.write(' topic=%r' % self.topic)
+        buf.write('>')
+        return buf.getvalue()
