@@ -3,7 +3,7 @@
    zeronimo.exceptions
    ~~~~~~~~~~~~~~~~~~~
 
-   :copyright: (c) 2013-2016 by Heungsub Lee
+   :copyright: (c) 2013-2017 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
 
 """
@@ -22,14 +22,6 @@ class ZeronimoException(BaseException):
     pass
 
 
-class Reject(ZeronimoException):
-    """Worker rejects a task.  Mark a method with ``defer_ack=True`` to control
-    rejection by this.
-    """
-
-    pass
-
-
 class EmissionError(ZeronimoException):
 
     pass
